@@ -10,7 +10,7 @@ import numpy
 import theano
 
 from theanolm import Network
-from theanolm.backend import TextFileType
+from theanolm.backend import TextFileType, get_default_device
 from theanolm.parsing import ScoringBatchIterator
 from theanolm.scoring import TextScorer
 
@@ -38,6 +38,7 @@ def add_arguments(parser):
     argument_group = parser.add_argument_group("scoring")
     argument_group.add_argument(
         '--output', metavar='DETAIL', type=str, default='perplexity',
+        choices=['perplexity', 'utterance-scores', 'word-scores'],
         help='what to output, one of "perplexity", "utterance-scores", '
              '"word-scores" (default "perplexity")')
     argument_group.add_argument(
@@ -49,6 +50,7 @@ def add_arguments(parser):
         help="exclude <unk> tokens from perplexity computation")
     argument_group.add_argument(
         '--subwords', metavar='MARKING', type=str, default=None,
+        choices=['word-boundary', 'prefix-affix', None],
         help='the subword vocabulary uses MARKING to indicate how words are '
              'formed from subwords; one of "word-boundary" (<w> token '
              'separates words), "prefix-affix" (subwords that can be '
@@ -59,12 +61,18 @@ def add_arguments(parser):
              'words according to their unigram frequencies in the training '
              'data')
 
+    argument_group = parser.add_argument_group("configuration")
+    argument_group.add_argument(
+        '--default-device', metavar='DEVICE', type=str, default=None,
+        help='when multiple GPUs are present, use DEVICE as default')
+
     argument_group = parser.add_argument_group("logging and debugging")
     argument_group.add_argument(
         '--log-file', metavar='FILE', type=str, default='-',
         help='path where to write log file (default is standard output)')
     argument_group.add_argument(
         '--log-level', metavar='LEVEL', type=str, default='info',
+        choices=['debug', 'info', 'warn'],
         help='minimum level of events to log, one of "debug", "info", "warn" '
              '(default "info")')
     argument_group.add_argument(
@@ -94,20 +102,21 @@ def score(args):
 
     if args.debug:
         theano.config.compute_test_value = 'warn'
-        print("Enabled computing test values for tensor variables.")
-        print("Warning: GpuArray backend will fail random number generation!")
+        logging.info("Enabled computing test values for tensor variables.")
+        logging.warning("GpuArray backend will fail random number generation!")
     else:
         theano.config.compute_test_value = 'off'
     theano.config.profile = args.profile
     theano.config.profile_memory = args.profile
 
-    network = Network.from_file(args.model_path, exclude_unk=args.exclude_unk)
+    default_device = get_default_device(args.default_device)
+    network = Network.from_file(args.model_path, exclude_unk=args.exclude_unk,
+                                default_device=default_device)
 
-    print("Building text scorer.")
-    sys.stdout.flush()
+    logging.info("Building text scorer.")
     scorer = TextScorer(network, args.shortlist, args.exclude_unk, args.profile)
 
-    print("Scoring text.")
+    logging.info("Scoring text.")
     if args.output == 'perplexity':
         _score_text(args.input_file, network.vocabulary, scorer,
                     args.output_file, args.log_base, args.subwords, False)
@@ -409,14 +418,13 @@ def _score_utterances(input_file, vocabulary, scorer, output_file,
         lm_score /= log_scale
         output_file.write(str(lm_score) + '\n')
         if (line_num + 1) % 1000 == 0:
-            print("{0} sentences scored.".format(line_num + 1))
-        sys.stdout.flush()
+            logging.info("%d sentences scored.", line_num + 1)
 
     if scorer.num_words == 0:
-        print("The input file contains no words.")
+        logging.info("The input file contains no words.")
     else:
-        print("{0} words processed, including start-of-sentence and "
-              "end-of-sentence tags, and {1} ({2:.1f} %) out-of-vocabulary "
+        logging.info("%d words processed, including start-of-sentence and "
+              "end-of-sentence tags, and %d (%.1f %%) out-of-vocabulary "
               "words".format(scorer.num_words,
                              scorer.num_unks,
                              scorer.num_unks / scorer.num_words))
