@@ -40,7 +40,7 @@ class GRULayer(BasicLayer):
                           scale=0.01, count=num_gates+1)
         # hidden state input weights for each gate and the candidate state
         self._init_weight('step_input/W', (output_size, output_size),
-                          count=num_gates+1)
+                          count=num_gates+1, split_to_devices=False)
         # biases for each gate and the candidate state
         self._init_bias('layer_input/b', output_size, [None] * (num_gates + 1))
 
