@@ -5,4 +5,5 @@
 
 from theanolm.scoring.textscorer import TextScorer
 from theanolm.scoring.latticedecoder import LatticeDecoder
-from theanolm.scoring.slflattice import SLFLattice
+from theanolm.scoring.latticebatch import LatticeBatch
+from theanolm.scoring.rescoredlattice import RescoredLattice
