@@ -3,7 +3,6 @@
 """A module that implements the LatticeDecoder class.
 """
 
-from copy import deepcopy
 import logging
 import math
 
@@ -26,9 +25,11 @@ class LatticeDecoder(object):
         A token represents a partial path through a word lattice. The decoder
         propagates a set of tokens through the lattice by
         """
+        __slots__ = ("history", "state", "ac_logprob", "lat_lm_logprob",
+                     "nn_lm_logprob", "recombination_hash", "total_logprob")
 
         def __init__(self,
-                     history=None,
+                     history=(),
                      state=None,
                      ac_logprob=logprob_type(0.0),
                      lat_lm_logprob=logprob_type(0.0),
@@ -62,13 +63,12 @@ class LatticeDecoder(object):
                                   lattice links
             """
 
-            self.history = [] if history is None else history
+            self.history = history
             self.state = [] if state is None else state
             self.ac_logprob = ac_logprob
             self.lat_lm_logprob = lat_lm_logprob
             self.nn_lm_logprob = nn_lm_logprob
             self.recombination_hash = None
-            self.lm_logprob = None
             self.total_logprob = None
 
         @classmethod
@@ -88,7 +88,7 @@ class LatticeDecoder(object):
             :returns: a copy of ``token``
             """
 
-            return cls(deepcopy(token.history),
+            return cls(token.history,
                        token.state,
                        token.ac_logprob,
                        token.lat_lm_logprob,
@@ -107,16 +107,17 @@ class LatticeDecoder(object):
                 limited_history = self.history
             else:
                 limited_history = self.history[-recombination_order:]
-            self.recombination_hash = hash(tuple(limited_history))
+            self.recombination_hash = hash(limited_history)
 
         def recompute_total(self, nn_lm_weight, lm_scale, wi_penalty,
                             linear=False):
             """Computes the interpolated language model log probability and
             the total log probability.
 
-            The interpolated LM log probability is saved in ``self.lm_logprob``.
-            The total log probability is computed by applying LM scale factor
-            and adding the acoustic log probability and word insertion penalty.
+            First the interpolated LM log probability is computed. Then the
+            total log probability is computed by applying the LM scale factor,
+            and adding the acoustic log probability and the word insertion
+            penalty.
 
             :type nn_lm_weight: logprob_type
             :param nn_lm_weight: weight of the neural network LM probability
@@ -136,16 +137,17 @@ class LatticeDecoder(object):
             """
 
             if linear:
-                self.lm_logprob = interpolate_linear(
+                lm_logprob = interpolate_linear(
                     self.nn_lm_logprob, self.lat_lm_logprob,
                     nn_lm_weight)
             else:
-                self.lm_logprob = interpolate_loglinear(
+                lm_logprob = interpolate_loglinear(
                     self.nn_lm_logprob, self.lat_lm_logprob,
                     nn_lm_weight, (1.0 - nn_lm_weight))
+
             self.total_logprob = self.ac_logprob
-            self.total_logprob += self.lm_logprob * lm_scale
-            self.total_logprob += wi_penalty * len(self.history)
+            self.total_logprob += lm_logprob * lm_scale
+            self.total_logprob += wi_penalty * (len(self.history) - 1)
 
         def history_words(self, vocabulary):
             """Converts the word IDs in the history to words using
@@ -250,6 +252,19 @@ class LatticeDecoder(object):
           number of words to consider when deciding whether two tokens should be
           recombined, or ``None`` for the entire word history
 
+        prune_extra_limit : float
+          if set, adjust the beam and max_tokens_per_node pruning relative to
+          the number of tokens; the limits are divided by the number of tokens
+          and multiplied by this factor
+
+        abs_min_beam : float
+          when using prune_extra_limit, this is the minimum that the beam will
+          be adjusted to
+
+        abs_min_max_tokens : float
+          when using prune_extra_limit, this is the minimum that the maximum
+          number of tokens will be adjusted to
+
         :type network: Network
         :param network: the neural network object
 
@@ -273,6 +288,11 @@ class LatticeDecoder(object):
         if self._beam is not None:
             self._beam = logprob_type(self._beam)
         self._recombination_order = decoding_options['recombination_order']
+        self._prune_extra_limit = decoding_options.get('prune_extra_limit', None)
+        self._abs_min_beam = decoding_options.get('abs_min_beam', 0)
+        self._abs_min_max_tokens = decoding_options.get('abs_min_max_tokens', 0)
+        if self._prune_extra_limit is None:
+            self._abs_min_beam = self._abs_min_max_tokens = 0
 
         if decoding_options['use_shortlist'] and \
            self._vocabulary.has_unigram_probs():
@@ -303,22 +323,26 @@ class LatticeDecoder(object):
             profile=profile,
             on_unused_input='ignore')
 
-        self._tokens = None
-        self._sorted_nodes = None
-
     def decode(self, lattice):
         """Propagates tokens through given lattice and returns a list of tokens
-        in the final node.
+        in the final nodes.
 
         Propagates tokens at a node to every outgoing link by creating a copy of
         each token and updating the language model scores according to the link.
 
+        The function returns two lists. The first list contains the final
+        tokens, sorted in the descending order of total log probability. I.e.
+        the first token in the list represents the best path through the
+        lattice. The second list contains the tokens that were dropped during
+        recombination. This is needed for constructing a new rescored lattice.
+
         :type lattice: Lattice
         :param lattice: a word lattice to be decoded
 
-        :rtype: list of LatticeDecoder.Tokens
-        :returns: the final tokens sorted by total log probability in descending
-                  order
+        :rtype: a tuple of two lists of LatticeDecoder.Tokens
+        :returns: a list of the final tokens sorted by probability (most likely
+                  token first), and a list of the tokens that were dropped
+                  during recombination
         """
 
         if self._lm_scale is not None:
@@ -330,60 +354,61 @@ class LatticeDecoder(object):
 
         if self._wi_penalty is not None:
             wi_penalty = logprob_type(self._wi_penalty)
-        if lattice.wi_penalty is not None:
+        elif lattice.wi_penalty is not None:
             wi_penalty = logprob_type(lattice.wi_penalty)
         else:
             wi_penalty = logprob_type(0.0)
 
-        self._tokens = [list() for _ in lattice.nodes]
+        tokens = [list() for _ in lattice.nodes]
+        recomb_tokens = []
         initial_state = RecurrentState(self._network.recurrent_state_size)
-        initial_token = self.Token(history=[self._sos_id], state=initial_state)
+        initial_token = self.Token(history=(self._sos_id,), state=initial_state)
         initial_token.recompute_hash(self._recombination_order)
         initial_token.recompute_total(self._nnlm_weight, lm_scale, wi_penalty,
                                       self._linear_interpolation)
-        self._tokens[lattice.initial_node.id].append(initial_token)
+        tokens[lattice.initial_node.id].append(initial_token)
         lattice.initial_node.best_logprob = initial_token.total_logprob
 
-        self._sorted_nodes = lattice.sorted_nodes()
-        nodes_processed = 0
-        for node in self._sorted_nodes:
-            node_tokens = self._tokens[node.id]
-            assert node_tokens
-            num_pruned_tokens = len(node_tokens)
-            self._prune(node)
-            node_tokens = self._tokens[node.id]
-            assert node_tokens
-            num_pruned_tokens -= len(node_tokens)
-
-            if node.id == lattice.final_node.id:
-                new_tokens = self._propagate(
-                    node_tokens, None, lm_scale, wi_penalty)
-                return sorted(new_tokens,
-                              key=lambda token: token.total_logprob,
-                              reverse=True)
+        sorted_nodes = lattice.sorted_nodes()
+        self._nodes_processed = 0
+        final_tokens = []
+        for node in sorted_nodes:
+            stats = self._prune(node, sorted_nodes, tokens, recomb_tokens)
 
             num_new_tokens = 0
+            node_tokens = tokens[node.id]
+            assert node_tokens
+            if node.final:
+                new_tokens = self._propagate(
+                    node_tokens, None, lm_scale, wi_penalty)
+                final_tokens.extend(new_tokens)
+                num_new_tokens += len(new_tokens)
             for link in node.out_links:
                 new_tokens = self._propagate(
                     node_tokens, link, lm_scale, wi_penalty)
-                self._tokens[link.end_node.id].extend(new_tokens)
+                tokens[link.end_node.id].extend(new_tokens)
+                # If there are lots of tokens in the end node, prune already to
+                # conserve memory.
+                if self._max_tokens_per_node is not None and \
+                   len(tokens[link.end_node.id]) > self._max_tokens_per_node * 2:
+                    self._prune(link.end_node, sorted_nodes, tokens, recomb_tokens)
                 num_new_tokens += len(new_tokens)
+            stats['new'] = num_new_tokens
 
-            nodes_processed += 1
-            if nodes_processed % math.ceil(len(self._sorted_nodes) / 20) == 0:
-                logging.debug("[%d] (%.2f %%) -- tokens = %d +%d -%d",
-                              nodes_processed,
-                              nodes_processed / len(self._sorted_nodes) * 100,
-                              len(node_tokens),
-                              num_new_tokens,
-                              num_pruned_tokens)
+            self._nodes_processed += 1
+            self._log_stats(stats, node.id, len(sorted_nodes))
 
-        raise InputError("Could not reach the final node of word lattice.")
+        if len(final_tokens) == 0:
+            raise InputError("Could not reach a final node of word lattice.")
+
+        final_tokens = self._sorted_recombined_tokens(final_tokens,
+                                                      recomb_tokens)
+        return final_tokens, recomb_tokens
 
     def _propagate(self, tokens, link, lm_scale, wi_penalty):
         """Propagates tokens to given link or to end of sentence.
 
-        Lattices may contain !NULL, !ENTER, !EXIT, etc. nodes that model e.g.
+        Lattices may contain null nodes with word ``None`` that model e.g.
         silence or sentence start or end, or for example when the topology is
         easier to represent with extra nodes. Such null nodes may contain
         language model scores. Then the function will update the acoustic and
@@ -421,7 +446,8 @@ class LatticeDecoder(object):
                     token.ac_logprob += link.ac_logprob
                 if link.lm_logprob is not None:
                     token.lat_lm_logprob += link.lm_logprob
-            if not link.word.startswith('!'):
+
+            if link.word is not None:
                 try:
                     word = self._vocabulary.word_to_id[link.word]
                 except KeyError:
@@ -437,59 +463,139 @@ class LatticeDecoder(object):
             token.recompute_hash(self._recombination_order)
             token.recompute_total(self._nnlm_weight, lm_scale, wi_penalty,
                                   self._linear_interpolation)
-            if link is not None:
+            if link is not None and link.end_node is not None:
                 if (link.end_node.best_logprob is None) or \
                    (token.total_logprob > link.end_node.best_logprob):
                     link.end_node.best_logprob = token.total_logprob
 
         return new_tokens
 
-    def _prune(self, node):
+    def _prune(self, node, sorted_nodes, tokens, recomb_tokens):
         """Prunes tokens from a node according to beam and the maximum number of
-        tokens.
+        tokens, and recombines tokens whose N previous words are identical
+        (where N is the recombination order).
 
         :type node: Lattice.Node
         :param node: perform pruning on this node
+
+        :type sorted_nodes: list of Lattice.Nodes
+        :param sorted_nodes: all nodes in topological order
+
+        :type tokens: list of LatticeDecoder.Tokens
+        :param tokens: all tokens
+
+        :type recomb_tokens: list of tuples
+        :param recomb_tokens: for all tokens that were dropped during
+            recombination, contains the token, and the history and NNLM log
+            probability of the kept token; dropped tokens will be added to the
+            list
+
+        :rtype: dict
+        :returns: a dictionary of statistics collected during pruning
         """
 
-        new_tokens = dict()
-        for token in self._tokens[node.id]:
-            key = token.recombination_hash
-            if (key not in new_tokens) or \
-               (token.total_logprob > new_tokens[key].total_logprob):
-                new_tokens[key] = token
+        node_tokens = tokens[node.id]
+        assert node_tokens
+        stats = dict()
+        stats['before'] = len(node_tokens)
 
-        # Sort the tokens by descending log probability.
-        new_tokens = sorted(new_tokens.values(),
-                            key=lambda token: token.total_logprob, reverse=True)
+        limit_divider = 1
+        if self._prune_extra_limit is not None:
+            limit_divider = len(sorted_nodes) // self._prune_extra_limit + 1
+            limit_divider = max(1, limit_divider)
+
+        new_tokens = self._sorted_recombined_tokens(node_tokens,
+                                                    recomb_tokens)
+        stats['after-recomb'] = len(new_tokens)
 
         # Compare to the best probability at the same or later time.
         if self._beam is not None:
             if node.time is None:
-                node_ids = [iter_node.id for iter_node in self._sorted_nodes]
+                node_ids = [iter_node.id for iter_node in sorted_nodes]
                 time_begin = node_ids.index(node.id)
             else:
-                for time_begin, iter_node in enumerate(self._sorted_nodes):
+                for time_begin, iter_node in enumerate(sorted_nodes):
                     if (iter_node.time is not None) and \
                        (iter_node.time >= node.time):
                         break
-            assert time_begin < len(self._sorted_nodes)
-
+            assert time_begin < len(sorted_nodes)
             best_logprob = max(iter_node.best_logprob
-                               for iter_node in self._sorted_nodes[time_begin:]
+                               for iter_node in sorted_nodes[time_begin:]
                                if iter_node.best_logprob is not None)
-            threshold = best_logprob - self._beam
-            token_index = len(new_tokens) - 1
-            while (token_index >= 1) and \
-                  (new_tokens[token_index].total_logprob <= threshold):
-                del new_tokens[token_index]
-                token_index -= 1
+
+            beam = self._beam / limit_divider
+            beam = max(beam, self._abs_min_beam)
+            threshold = best_logprob - beam
+
+            stats['best'] = best_logprob
+            stats['node-best'] = new_tokens[0].total_logprob
+            stats['node-worst'] = new_tokens[-1].total_logprob
+            stats['threshold'] = threshold
+            stats['average'] = sum(t.total_logprob for t in new_tokens) / len(new_tokens)
+            if len(new_tokens) >= 10:
+                stats['pos10'] = new_tokens[9].total_logprob
+            if len(new_tokens) >= 50:
+                stats['pos50'] = new_tokens[49].total_logprob
+            if len(new_tokens) >= 100:
+                stats['pos100'] = new_tokens[99].total_logprob
+
+            keep_tokens = len(new_tokens)
+            while (keep_tokens > 1) and \
+                  (new_tokens[keep_tokens - 1].total_logprob <= threshold):
+                keep_tokens -= 1
+            new_tokens = new_tokens[:keep_tokens]
+        stats['after-beam'] = len(new_tokens)
 
         # Enforce limit on number of tokens at each node.
         if self._max_tokens_per_node is not None:
-            new_tokens[self._max_tokens_per_node:] = []
+            max_tokens = self._max_tokens_per_node // limit_divider
+            max_tokens = max(max_tokens, self._abs_min_max_tokens)
+            new_tokens = new_tokens[:max_tokens]
+        stats['after-max'] = len(new_tokens)
 
-        self._tokens[node.id] = new_tokens
+        tokens[node.id] = new_tokens
+        return stats
+
+    def _sorted_recombined_tokens(self, tokens, recomb_tokens):
+        """Sorts tokens by descending probability and recombines tokens with
+        identical hash.
+
+        First sorts the tokens by probability, most likely token first. Then
+        keeps only the first if there are multiple tokens with the same hash.
+        Tokens with identical N previous words in the history (where N is the
+        recombination order) will be dropped and added to recomb_tokens.
+
+        :type tokens: list of LatticeDecoder.Tokens
+        :param tokens: input tokens
+
+        :type recomb_tokens: list of tuples
+        :param recomb_tokens: for all tokens that were dropped during
+            recombination, contains the token, and the history and NNLM log
+            probability of the kept token; dropped tokens will be added to the
+            list
+
+        :rtype: list of LatticeDecoder.Tokens
+        :returns: the tokens in sorted order and with only the best token of
+                  those with identical recombination hash
+        """
+
+        sorted_tokens = sorted(tokens,
+                               key=lambda token: token.total_logprob,
+                               reverse=True)
+
+        token_map = dict()
+        result = []
+        for token in sorted_tokens:
+            key = token.recombination_hash
+            if key not in token_map:
+                result.append(token)
+                token_map[key] = token
+            else:
+                kept_token = token_map[key]
+                recomb_tokens.append((token,
+                                      kept_token.history,
+                                      kept_token.nn_lm_logprob))
+        return result
 
     def _append_word(self, tokens, target_word, oov_logprob=None):
         """Appends a word to each of the given tokens, and updates their scores.
@@ -537,10 +643,10 @@ class LatticeDecoder(object):
         output_state = step_result[1:]
 
         for index, token in enumerate(tokens):
-            token.history.append(target_word)
+            token.history = token.history + (target_word,)
             token.state = RecurrentState(self._network.recurrent_state_size)
             # Slice the sequence that corresponds to this token.
-            token.state.set([layer_state[:, index:index+1]
+            token.state.set([layer_state[:, index:index + 1]
                              for layer_state in output_state])
             # logprobs matrix contains only one time step.
             token.nn_lm_logprob += self._handle_unk_logprob(target_word,
@@ -589,3 +695,35 @@ class LatticeDecoder(object):
             return oov_logprob
 
         return network_logprob
+
+    def _log_stats(self, stats, node_id, num_nodes):
+        """Writes pruning statistics to debug log.
+
+        :type stats: dict
+        :param stats: a dictionary of statistics on the number of tokens and
+                      their log probabilities
+        """
+
+        if self._nodes_processed % math.ceil(num_nodes / 20) != 0:
+            return
+
+        optional = ''
+        for name in ('after-recomb', 'after-beam', 'after-max', 'new'):
+            if name in stats:
+                optional += ' {}={}'.format(name, stats[name])
+        logging.debug('[%d] (%.2f %%) node=%d -- tokens before=%d%s',
+                      self._nodes_processed,
+                      self._nodes_processed / num_nodes * 100,
+                      node_id,
+                      stats['before'],
+                      optional)
+
+        if 'best' in stats:
+            optional = ''
+            for name in ('node-best', 'node-worst', 'threshold', 'average',
+                         'pos10', 'pos50', 'pos100'):
+                if name in stats:
+                    optional += ' {}={:.1f}'.format(name, stats[name])
+            logging.debug('logprob best=%.1f%s',
+                          stats['best'],
+                          optional)
